<?php
namespace metagen_php;
use Exception;

function codegen(?string $cache_dir, \mtgMetaInfo $meta, array $options = []) : string
{
  $twig = get_twig();

  if(!empty($cache_dir))
    $twig->setCache($cache_dir);

  $options['meta'] = $meta;

  return $twig->render('codegen_bundle.twig', $options);
}

function get_twig(array $inc_path = [])
{
  array_unshift($inc_path, __DIR__ . "/../tpl/");
  $loader = new \Twig\Loader\FilesystemLoader($inc_path);

  $twig = new \Twig\Environment($loader, [
    'debug' => true, 
    'autoescape' => false, 
    'strict_variables' => true]
  );
  $twig->addExtension(new \Twig\Extension\DebugExtension());

  _add_twig_support($twig);

  return $twig;
}

function supported_tokens()
{
  return [
    'default',
    'virtual',
  ];
}

function _add_twig_support(\Twig\Environment $twig)
{
  $twig->addTest(new \Twig\TwigTest('instanceof', 
    function($obj, $class) 
    { 
      return (new \ReflectionClass($class))->isInstance($obj);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('var_export', 
    function($v) 
    { 
      return str_replace("\n", "", var_export($v, true));
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('Error', 
    function($e) 
    {
      throw new Exception($e);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('has_token', 
    function($o, $token) 
    { 
      return $o->hasToken($token);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('token', 
    function($o, $name) 
    {
      return $o->getToken($name);
    }
  ));
  $twig->addFilter(new \Twig\TwigFilter('default_value', 
    function($default) 
    { 
      return default_value($default);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('apply_value_filters', 
    function ($field_name, array $tokens, $val, $add_assoc_check = true)
    {
      return apply_value_filters($field_name, $tokens, $val, $add_assoc_check);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('data2value', 
    function($val, \mtgType $type, $data, $tokens) 
    { 
      return data2value($val, $type, $data, $tokens);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('value2data', 
    function($val, $name, \mtgType $type, $data, $tokens = array()) 
    { 
      return value2data($val, $name, $type, $data, $tokens);
    }
  ));
  $twig->addFunction(new \Twig\TwigFunction('data_normalize', 
    function($name, \mtgType $type, $buf, $prefix = '', $tokens = array()) 
    { 
      return data_normalize($name, $type, $buf, $prefix, $tokens);
    }
  ));
}

function default_value($default)
{
  if($default == "[]")
    return 'array()';

  if(strlen($default) > 2 && strpos($default, '[') === 0 && strpos($default, ']') === strlen($default)-1)
  {
    $str = trim($default, '][');
    if(strpos($str, "{") !== false )
      $str = var_export(json_decode(trim($default, ']['), true), true);
    return 'array(' . $str . ')';
  }
  return str_replace('$', '\$', $default);
}

function apply_filters($token_prefix, $field_name, array $tokens, $val, $add_assoc_check = true)
{
  $str = $val;
  foreach($tokens as $token => $args_json)
  {
    if(strpos($token, $token_prefix) === false)
      continue;

    $filter_func = '\metagen_php\\' . $token; 

    if(function_exists($filter_func))
    {
      $args = null;
      if($args_json && 
        ((strpos($args_json, '[') === 0 && strlen($args_json) > 1 && $args_json[strlen($args_json)-1] === ']') || 
         (strpos($args_json, '{') === 0 && strlen($args_json) > 1 && $args_json[strlen($args_json)-1] === '}')))
      {
        $args = json_decode($args_json, true);
        if(!is_array($args))
          throw new Exception("Bad filter '$token' args: '$args_json'");
      }
      else
        $args = $args_json;

      if($add_assoc_check)
        $str = "(\$assoc ? $filter_func($str, '$field_name', \$this, " . str_replace("\n", "",  var_export($args, true)) . ") : $str)";
      else
        $str = "$filter_func($str, '$field_name', \$this, " . str_replace("\n", "",  var_export($args, true)) . ")";
    }
    else
      throw new Exception("No such filter '$filter_func'");
  }
  return $str;
}

function apply_filters_v2($token_prefix, $field_name, array $tokens, $val) : string
{
  $str = $val;
  foreach($tokens as $token => $args_json)
  {
    if(strpos($token, $token_prefix) === false)
      continue;

    $filter_func = '\metagen_php\\' . $token;

    if(function_exists($filter_func))
    {
      $args = null;
      if($args_json && 
        ((strpos($args_json, '[') === 0 && strlen($args_json) > 1 && $args_json[strlen($args_json)-1] === ']') || 
         (strpos($args_json, '{') === 0 && strlen($args_json) > 1 && $args_json[strlen($args_json)-1] === '}')))
      {
        $args = json_decode($args_json, true);
        if(!is_array($args))
          throw new Exception("Bad filter '$token' args: '$args_json'");
      }
      else
        $args = $args_json;

      $str = "$filter_func($str, '$field_name', \$data, " . str_replace("\n", "",  var_export($args, true)) . ")";
    }
    else
      throw new Exception("No such filter '$filter_func'");
  }
  if($str !== $val)
    return "{$val} = " . $str; 
  else
    return '';
}

function apply_array_filters($field_name, array $tokens, $val, $add_assoc_check = true)
{
  return apply_filters("fltarr_", $field_name, $tokens, $val, $add_assoc_check);
}

function apply_value_filters($field_name, array $tokens, $val, $add_assoc_check = true)
{
  return apply_filters("flt_", $field_name, $tokens, $val, $add_assoc_check);
}

function apply_array_filters_v2($field_name, array $tokens, $val) : string
{
  return apply_filters_v2("fltarr_", $field_name, $tokens, $val);
}

function apply_value_filters_v2($field_name, array $tokens, $val) : string
{
  return apply_filters_v2("flt_", $field_name, $tokens, $val);
}

//TODO: move it to template
function data2value(string $val, \mtgType $type, string $data, array $tokens, bool $inside_loop = false)
{
  $str = '';
  $tmp_val = '$tmp_val__';

  if(!$inside_loop)
  {
    $str .= "{$tmp_val} = current({$data});\n";
    $str .= "if({$tmp_val} === false && key({$data}) === null) throw new Exception(\"No more data\");\n";
    $str .= "next({$data});\n\n";
  }

  if($type instanceof \mtgBuiltinType || $type instanceof \mtgMetaEnum)
  {
    $str .= "{$val} = {$tmp_val};\n";
  }
  else if($type instanceof \mtgMetaStruct)
  {
    $str .= "if({$tmp_val} === null)\n";
    $str .= "  {$val} = null;\n";
    $str .= "else\n";
    $str .= "{\n";
    if(array_key_exists('virtual', $tokens))
    {
      $str .= "  \$vclass__ = AutogenBundle::getClassName(array_shift({$tmp_val}));\n";
      $str .= "  if(!is_a(\$vclass__, '{$type->getName()}', true)) throw new Exception(\"'\$vclass__' is not subclass of '{$type->getName()}'\");\n";
      $str .= "  {$val} = new \$vclass__();\n";
      $str .= "  {$val}->import({$tmp_val});\n";
    }
    else
    {
      $str .= "  {$val} = new {$type}();\n";
      $str .= "  {$val}->import({$tmp_val});\n";
    }
    $str .= "}\n";
  }
  else if($type instanceof \mtgArrType)
  {
    $str .= "\$tmp_arr__ = {$tmp_val};\n";
    $str .= "foreach(\$tmp_arr__ as {$tmp_val})\n";
    $str .= "{\n";
    $str .= data2value('$tmp__', $type->getValue(), "\$tmp_arr_item__", $tokens, true) . "\n";
    $str .= "{$val}[] = \$tmp__;\n";
    $str .= "}\n";
  }
  else
    throw new Exception("Unknown type '{$type->getName()}'");

  return $str;
}

//TODO: move it to template
function value2data(string $val, string $name, \mtgType $type, string $data, $tokens = array(), $assoc = '$assoc')
{
  $str = '';

  if($type instanceof \mtgBuiltinType)
  {
    $str .= "\metagen_php\array_set_value({$data}, {$assoc}, '$name', $val);\n";
  }
  else if($type instanceof \mtgMetaStruct)
  {
    if(array_key_exists('virtual', $tokens))
      $str .= "\metagen_php\array_set_value({$data}, {$assoc}, '$name', {$val}->export(\$assoc, true/*virtual*/));\n";
    else
      $str .= "\metagen_php\array_set_value({$data}, {$assoc}, '$name',  {$val}->export(\$assoc));\n";
  }
  else if($type instanceof \mtgMetaEnum)
  {
    $str .= "\metagen_php\array_set_value({$data}, {$assoc}, '$name', $val);\n";
  }
  else if($type instanceof \mtgArrType)
  {
    $str .= "\$arr_tmp__ = array();\n";
    //NOTE: adding optimization, checking if the first item is array
    $str .= "if({$val} && is_array(current({$val})))\n";
    $str .= "  \$arr_tmp__ = {$val};\n";
    $str .= "else\n";
    $str .= "  foreach({$val} as \$arr_tmp_item__)\n";
    $str .= "  {\n";
    $str .= value2data("\$arr_tmp_item__", '', $type->getValue(), "\$arr_tmp__", $tokens, 'false') . "\n";
    $str .= "  }\n";
    $str .= "\metagen_php\array_set_value({$data}, \$assoc, '$name', \$arr_tmp__);\n";
  }
  else
    throw new Exception("Unknown type '{$type->getName()}'");

  return $str;
}

function is_null_str($default)
{
  return is_string($default) && json_decode($default, true) === null;
}

function get_default_value_arg(\mtgType $type, array $tokens, bool &$has_default = true) : string
{
  $has_default = array_key_exists('default', $tokens);
  $default_value_arg =  $has_default ? default_value($tokens['default']) : "null";

  if($type instanceof \mtgMetaStruct)
  {
    $default = array_key_exists('default', $tokens) ? $tokens['default'] : null;
    if($default)
    {
      $default = json_decode($default, true);
      if(is_array($default))
      {
        $default = str_replace("\n", "",  var_export($default, true));
        $default_value_arg = "$default";
      }
      else if($default === null)
      {
        $default_value_arg = "null";
      }
      else
        throw new Exception("Bad default struct: " . $tokens['default']);
    }
  }

  return $default_value_arg;
}

function data_normalize($name, \mtgType $type, $buf, $dst, $tokens = array(), $tmp_val = '$tmp_val__')
{
  $str = '';

  $has_default = false;
  $default_value_arg = get_default_value_arg($type, $tokens, $has_default);

  $str .= "\n";

  if($name !== null)
  {
    $str .= "{$tmp_val} = {$buf}['{$name}'] ?? {$default_value_arg};\n";
    if(!$has_default)
      $str .= "if({$tmp_val} === null) throw new Exception(\"'{$name}' is missing\");\n";
    $str .= "if(\$check_junk) unset({$buf}['{$name}']);\n";
  }

  if($type instanceof \mtgBuiltinType)
  {
    $str .= apply_value_filters_v2($name, $tokens, "{$tmp_val}"). ";\n";
    $str .= "{$dst}[] = \metagen_php\\val_{$type}({$tmp_val});\n";
  }
  else if($type instanceof \mtgMetaStruct)
  {
    $str .= "if({$tmp_val} === null) {\n";
    $str .= "  {$dst}[] = null; \n";
    $str .= "} else {\n";
    $str .= apply_value_filters_v2($name, $tokens, "{$tmp_val}"). ";\n";
    $str .= "\$tmp_sub_arr__ = \metagen_php\\val_arr({$tmp_val});\n";
    if(array_key_exists('virtual', $tokens))
    {
      $str .= "\$vclass_id__ = \metagen_php\\val_uint32(\$tmp_sub_arr__['\$id'] ?? {$type->getClassId()});\n";
      $str .= "if(\$check_junk) unset(\$tmp_sub_arr__['\$id']);\n";
      $str .= "\$vclass__ = AutogenBundle::getClassName(\$vclass_id__);\n";
      $str .= "if(!is_a(\$vclass__, '{$type->getName()}', true)) throw new Exception(\"'\$vclass__' is not subclass of '{$type->getName()}'\");\n";
      $str .= "\$tmp_sub_mapped__ = array(\$vclass_id__);\n";
      $str .= "call_user_func_array([\$vclass__, 'normalize'], array(&\$tmp_sub_arr__, &\$tmp_sub_mapped__, true, \$check_junk));\n";
      $str .= "{$dst}[] = \$tmp_sub_mapped__;\n";
    }
    else
    {
      $str .= "\$tmp_sub_mapped__ = array();\n";
      $str .= "{$type}::normalize(\$tmp_sub_arr__, \$tmp_sub_mapped__, true, \$check_junk);\n";
      $str .= "{$dst}[] = \$tmp_sub_mapped__;\n";
    }
    $str .= "}";
  }
  else if($type instanceof \mtgArrType)
  {
    //TODO: maybe filters should be applied to the whole array as well?
    $str .= apply_array_filters_v2($name, $tokens, "{$tmp_val}"). ";\n";
    $str .= "\$tmp_arr__ = \metagen_php\\val_arr({$tmp_val});\n";
    $str .= "\$tmp_dst_arr__ = array();\n";

    $str .= "foreach(\$tmp_arr__ as \$tmp_arr_item__)\n";
    $str .= "{\n";
    //NOTE: removing default for field
    unset($tokens['default']);
    $str .= data_normalize(null, $type->getValue(), "\$tmp_arr_item__", "\$tmp_dst_arr__", $tokens, "\$tmp_arr_item__") . "\n";
    $str .= "}\n";
    $str .= "{$dst}[] = \$tmp_dst_arr__;\n";
  }
  else if($type instanceof \mtgMetaEnum)
  {
    $str .= apply_value_filters_v2($name, $tokens, "{$tmp_val}"). ";\n";
    $str .= "{$dst}[] = \metagen_php\\val_enum('$type', {$tmp_val}, false);\n";
  }
  else
    throw new Exception("Unknown type '{$type->getName()}'");

  return $str;
}

